@@ -1,17 +1,20 @@
 """ Header text for the C and Fortran BLAS interfaces.
 
 There is no standard name or location for this header, so we just insert it
-ourselves into the C code
+ourselves into the C code.
+
 """
+from __future__ import absolute_import, print_function, division
 import logging
 import textwrap
 import sys
+import os
+from os.path import dirname
 
 from theano import config
 from theano.gof.cmodule import GCC_compiler
 
 _logger = logging.getLogger('theano.tensor.blas')
-#_logger.setLevel(logging.INFO)
 
 
 def detect_macos_sdot_bug():
@@ -32,6 +35,7 @@ def detect_macos_sdot_bug():
           detected. Its value is returned by the function
         - detect_macos_sdot_bug.fix_works will be set to True if the fix was
           attempted, and succeeded.
+
     """
     _logger.debug('Starting detection of bug in Mac OS BLAS sdot_ routine')
     if detect_macos_sdot_bug.tested:
@@ -49,8 +53,17 @@ def detect_macos_sdot_bug():
         # Library directories should also be added as rpath,
         # so that they can be loaded even if the environment
         # variable LD_LIBRARY_PATH does not contain them
+        lib_path = os.environ.get('DYLD_FALLBACK_LIBRARY_PATH', '').split(':')
         if f.startswith('-L'):
             flags.append('-Wl,-rpath,' + f[2:])
+            # also append those paths to DYLD_FALLBACK_LIBRARY_PATH to
+            # support libraries that have the wrong install_name
+            # (such as MKL on canopy installs)
+            if (f[2:] not in lib_path):
+                lib_path.append(f[2:])
+        # this goes into the python process environment that is
+        # inherited by subprocesses/used by dyld when loading new objects
+        os.environ['DYLD_FALLBACK_LIBRARY_PATH'] = ':'.join(lib_path)
 
     test_code = textwrap.dedent("""\
         extern "C" float sdot_(int*, float*, int*, float*, int*);
@@ -70,8 +83,10 @@ def detect_macos_sdot_bug():
         """)
 
     _logger.debug('Trying to compile and run test case.')
-    compilation_ok, run_ok = GCC_compiler.try_compile_tmp(test_code,
-            tmp_prefix='detect_macos_sdot_bug_', flags=flags, try_run=True)
+    compilation_ok, run_ok = GCC_compiler.try_compile_tmp(
+        test_code,
+        tmp_prefix='detect_macos_sdot_bug_',
+        flags=flags, try_run=True)
     detect_macos_sdot_bug.tested = True
 
     # If compilation failed, we consider there is a bug,
@@ -92,9 +107,8 @@ def detect_macos_sdot_bug():
 
     # Then, try a simple fix
     test_fix_code = textwrap.dedent("""\
-        extern "C" float sdot_(int*, float*, int*, float*, int*);
         extern "C" float cblas_sdot(int, float*, int, float*, int);
-        float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
+        static float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
         {
             return cblas_sdot(*Nx, x, *Sx, y, *Sy);
         }
@@ -116,13 +130,13 @@ def detect_macos_sdot_bug():
 
     _logger.debug('Trying to compile and run tentative workaround.')
     compilation_fix_ok, run_fix_ok = GCC_compiler.try_compile_tmp(
-            test_fix_code,
-            tmp_prefix='detect_macos_sdot_bug_testfix_',
-            flags=flags,
-            try_run=True)
+        test_fix_code,
+        tmp_prefix='detect_macos_sdot_bug_testfix_',
+        flags=flags,
+        try_run=True)
 
     _logger.info("Status of tentative fix -- compilation OK: %s, works: %s",
-            compilation_fix_ok, run_fix_ok)
+                 compilation_fix_ok, run_fix_ok)
     detect_macos_sdot_bug.fix_works = run_fix_ok
 
     return detect_macos_sdot_bug.present
@@ -216,39 +230,39 @@ def cblas_header_text():
      * ===========================================================================
      */
 
-    /* 
+    /*
      * Routines with standard 4 prefixes (s, d, c, z)
      */
-    void cblas_sswap(const int N, float *X, const int incX, 
+    void cblas_sswap(const int N, float *X, const int incX,
                      float *Y, const int incY);
-    void cblas_scopy(const int N, const float *X, const int incX, 
+    void cblas_scopy(const int N, const float *X, const int incX,
                      float *Y, const int incY);
     void cblas_saxpy(const int N, const float alpha, const float *X,
                      const int incX, float *Y, const int incY);
 
-    void cblas_dswap(const int N, double *X, const int incX, 
+    void cblas_dswap(const int N, double *X, const int incX,
                      double *Y, const int incY);
-    void cblas_dcopy(const int N, const double *X, const int incX, 
+    void cblas_dcopy(const int N, const double *X, const int incX,
                      double *Y, const int incY);
     void cblas_daxpy(const int N, const double alpha, const double *X,
                      const int incX, double *Y, const int incY);
 
-    void cblas_cswap(const int N, void *X, const int incX, 
+    void cblas_cswap(const int N, void *X, const int incX,
                      void *Y, const int incY);
-    void cblas_ccopy(const int N, const void *X, const int incX, 
+    void cblas_ccopy(const int N, const void *X, const int incX,
                      void *Y, const int incY);
     void cblas_caxpy(const int N, const void *alpha, const void *X,
                      const int incX, void *Y, const int incY);
 
-    void cblas_zswap(const int N, void *X, const int incX, 
+    void cblas_zswap(const int N, void *X, const int incX,
                      void *Y, const int incY);
-    void cblas_zcopy(const int N, const void *X, const int incX, 
+    void cblas_zcopy(const int N, const void *X, const int incX,
                      void *Y, const int incY);
     void cblas_zaxpy(const int N, const void *alpha, const void *X,
                      const int incX, void *Y, const int incY);
 
 
-    /* 
+    /*
      * Routines with S and D prefix only
      */
     void cblas_srotg(float *a, float *b, float *c, float *s);
@@ -266,7 +280,7 @@ def cblas_header_text():
                     double *Y, const int incY, const double *P);
 
 
-    /* 
+    /*
      * Routines with S D C Z CS and ZD prefixes
      */
     void cblas_sscal(const int N, const float alpha, float *X, const int incX);
@@ -282,7 +296,7 @@ def cblas_header_text():
      * ===========================================================================
      */
 
-    /* 
+    /*
      * Routines with standard 4 prefixes (S, D, C, Z)
      */
     void cblas_sgemv(const enum CBLAS_ORDER order,
@@ -297,11 +311,11 @@ def cblas_header_text():
                      const int incX, const float beta, float *Y, const int incY);
     void cblas_strmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const float *A, const int lda, 
+                     const int N, const float *A, const int lda,
                      float *X, const int incX);
     void cblas_stbmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const int K, const float *A, const int lda, 
+                     const int N, const int K, const float *A, const int lda,
                      float *X, const int incX);
     void cblas_stpmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
@@ -330,11 +344,11 @@ def cblas_header_text():
                      const int incX, const double beta, double *Y, const int incY);
     void cblas_dtrmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const double *A, const int lda, 
+                     const int N, const double *A, const int lda,
                      double *X, const int incX);
     void cblas_dtbmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const int K, const double *A, const int lda, 
+                     const int N, const int K, const double *A, const int lda,
                      double *X, const int incX);
     void cblas_dtpmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
@@ -363,11 +377,11 @@ def cblas_header_text():
                      const int incX, const void *beta, void *Y, const int incY);
     void cblas_ctrmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const void *A, const int lda, 
+                     const int N, const void *A, const int lda,
                      void *X, const int incX);
     void cblas_ctbmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const int K, const void *A, const int lda, 
+                     const int N, const int K, const void *A, const int lda,
                      void *X, const int incX);
     void cblas_ctpmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
@@ -396,11 +410,11 @@ def cblas_header_text():
                      const int incX, const void *beta, void *Y, const int incY);
     void cblas_ztrmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const void *A, const int lda, 
+                     const int N, const void *A, const int lda,
                      void *X, const int incX);
     void cblas_ztbmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
-                     const int N, const int K, const void *A, const int lda, 
+                     const int N, const int K, const void *A, const int lda,
                      void *X, const int incX);
     void cblas_ztpmv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
                      const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_DIAG Diag,
@@ -418,7 +432,7 @@ def cblas_header_text():
                      const int N, const void *Ap, void *X, const int incX);
 
 
-    /* 
+    /*
      * Routines with S and D prefixes only
      */
     void cblas_ssymv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
@@ -480,7 +494,7 @@ def cblas_header_text():
                     const int incX, const double *Y, const int incY, double *A);
 
 
-    /* 
+    /*
      * Routines with C and Z prefixes only
      */
     void cblas_chemv(const enum CBLAS_ORDER order, const enum CBLAS_UPLO Uplo,
@@ -551,7 +565,7 @@ def cblas_header_text():
      * ===========================================================================
      */
 
-    /* 
+    /*
      * Routines with standard 4 prefixes (S, D, C, Z)
      */
     void cblas_sgemm(const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA,
@@ -675,7 +689,7 @@ def cblas_header_text():
                      void *B, const int ldb);
 
 
-    /* 
+    /*
      * Routines with prefixes C and Z only
      */
     void cblas_chemm(const enum CBLAS_ORDER Order, const enum CBLAS_SIDE Side,
@@ -716,6 +730,28 @@ def cblas_header_text():
 
 def blas_header_text():
     """C header for the fortran blas interface"""
+
+    blas_code = ""
+    if not config.blas.ldflags:
+        # Include the Numpy version implementation of [sd]gemm_.
+        current_filedir = dirname(__file__)
+        blas_common_filepath = os.path.join(current_filedir, 'c_code', 'alt_blas_common.h')
+        blas_template_filepath = os.path.join(current_filedir, 'c_code', 'alt_blas_template.c')
+        common_code = ""
+        sblas_code = ""
+        dblas_code = ""
+        with open(blas_common_filepath) as code:
+            common_code = code.read()
+        with open(blas_template_filepath) as code:
+            template_code = code.read()
+            sblas_code = template_code % {"float_type": "float", "float_size": 4, "npy_float": "NPY_FLOAT32", "precision": "s"}
+            dblas_code = template_code % {"float_type": "double", "float_size": 8, "npy_float": "NPY_FLOAT64", "precision": "d"}
+        if not common_code or not template_code:
+            raise IOError("Unable to load NumPy implementation of BLAS functions from C source files.")
+        blas_code += common_code
+        blas_code += sblas_code
+        blas_code += dblas_code
+
     header = """
     extern "C"
     {
@@ -729,7 +765,7 @@ def blas_header_text():
     /* Single Precision */
 
         void srot_(const int*, float *, const int*, float *, const int*, const float *, const float *);
-        void srotg_(float *,float *,float *,float *);    
+        void srotg_(float *,float *,float *,float *);
         void srotm_( const int*, float *, const int*, float *, const int*, const float *);
         void srotmg_(float *,float *,float *,const float *, float *);
         void sswap_( const int*, float *, const int*, float *, const int*);
@@ -746,7 +782,7 @@ def blas_header_text():
     /* Double Precision */
 
         void drot_(const int*, double *, const int*, double *, const int*, const double *, const double *);
-        void drotg_(double *,double *,double *,double *);    
+        void drotg_(double *,double *,double *,double *);
         void drotm_( const int*, double *, const int*, double *, const int*, const double *);
         void drotmg_(double *,double *,double *,const double *, double *);
         void dswap_( const int*, double *, const int*, double *, const int*);
@@ -808,8 +844,8 @@ def blas_header_text():
         void stpsv_( char*, char*, char*, const int*, const float *, float *, const int*);
         void sger_( const int*, const int*, const float *, const float *, const int*, const float *, const int*, float *, const int*);
         void ssyr_(char*, const int*, const float *, const float *, const int*, float *, const int*);
-        void sspr_(char*, const int*, const float *, const float *, const int*, float *); 
-        void sspr2_(char*, const int*, const float *, const float *, const int*, const float *, const int*,  float *); 
+        void sspr_(char*, const int*, const float *, const float *, const int*, float *);
+        void sspr2_(char*, const int*, const float *, const float *, const int*, const float *, const int*,  float *);
         void ssyr2_(char*, const int*, const float *, const float *, const int*, const float *, const int*,  float *, const int*);
 
     /* Double Precision */
@@ -827,8 +863,8 @@ def blas_header_text():
         void dtpsv_( char*, char*, char*, const int*, const double *, double *, const int*);
         void dger_( const int*, const int*, const double *, const double *, const int*, const double *, const int*, double *, const int*);
         void dsyr_(char*, const int*, const double *, const double *, const int*, double *, const int*);
-        void dspr_(char*, const int*, const double *, const double *, const int*, double *); 
-        void dspr2_(char*, const int*, const double *, const double *, const int*, const double *, const int*,  double *); 
+        void dspr_(char*, const int*, const double *, const double *, const int*, double *);
+        void dspr2_(char*, const int*, const double *, const double *, const int*, const double *, const int*,  double *);
         void dsyr2_(char*, const int*, const double *, const double *, const int*, const double *, const int*,  double *, const int*);
 
     /* Single Complex Precision */
@@ -924,7 +960,7 @@ def blas_header_text():
         if detect_macos_sdot_bug.fix_works:
             header += textwrap.dedent("""\
                     extern "C" float cblas_sdot(int, float*, int, float*, int);
-                    float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
+                    static float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
                     {
                         return cblas_sdot(*Nx, x, *Sx, y, *Sy);
                     }
@@ -932,7 +968,7 @@ def blas_header_text():
         else:
             # Make sure the buggy version of sdot_ is never used
             header += textwrap.dedent("""\
-                    float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
+                    static float sdot_(int* Nx, float* x, int* Sx, float* y, int* Sy)
                     {
                         fprintf(stderr,
                             "FATAL: The implementation of BLAS SDOT "
@@ -945,12 +981,59 @@ def blas_header_text():
                     }
                     """)
 
+    return header + blas_code
+
+
+if not config.blas.ldflags:
+    _logger.warning('Using NumPy C-API based implementation for BLAS functions.')
+
+
+def mkl_threads_text():
+    """C header for MKL threads interface"""
+    header = """
+    extern "C"
+    {
+        int     MKL_Set_Num_Threads_Local(int);
+        #define mkl_set_num_threads_local   MKL_Set_Num_Threads_Local
+
+        void    MKL_Set_Num_Threads(int);
+        #define mkl_set_num_threads         MKL_Set_Num_Threads
+
+        int     MKL_Get_Max_Threads(void);
+        #define mkl_get_max_threads         MKL_Get_Max_Threads
+
+        int     MKL_Domain_Set_Num_Threads(int, int);
+        #define mkl_domain_set_num_threads  MKL_Domain_Set_Num_Threads
+
+        int     MKL_Domain_Get_Max_Threads(int);
+        #define mkl_domain_get_max_threads  MKL_Domain_Get_Max_Threads
+
+        void    MKL_Set_Dynamic(int);
+        #define mkl_set_dynamic             MKL_Set_Dynamic
+
+        int     MKL_Get_Dynamic(void);
+        #define mkl_get_dynamic             MKL_Get_Dynamic
+    }
+    """
+    return header
+
+
+def openblas_threads_text():
+    """C header for OpenBLAS threads interface"""
+    header = """
+    extern "C"
+    {
+        void openblas_set_num_threads(int);
+        void goto_set_num_threads(int);
+        int openblas_get_num_threads(void);
+    }
+    """
     return header
 
 
 def blas_header_version():
     # Version for the base header
-    version = (1,)
+    version = (9,)
     if detect_macos_sdot_bug():
         if detect_macos_sdot_bug.fix_works:
             # Version with fix
@@ -988,15 +1071,15 @@ def ____gemm_code(check_ab, a_init, b_init):
 
         %(check_ab)s
 
-        if ((PyArray_DESCR(_x)->type_num != NPY_DOUBLE) 
+        if ((PyArray_DESCR(_x)->type_num != NPY_DOUBLE)
             && (PyArray_DESCR(_x)->type_num != NPY_FLOAT))
             goto _dot_execute_fallback;
 
-        if ((PyArray_DESCR(_y)->type_num != NPY_DOUBLE) 
+        if ((PyArray_DESCR(_y)->type_num != NPY_DOUBLE)
             && (PyArray_DESCR(_y)->type_num != NPY_FLOAT))
             goto _dot_execute_fallback;
 
-        if ((PyArray_DESCR(_y)->type_num != NPY_DOUBLE) 
+        if ((PyArray_DESCR(_y)->type_num != NPY_DOUBLE)
             && (PyArray_DESCR(_y)->type_num != NPY_FLOAT))
             goto _dot_execute_fallback;
 
@@ -1090,13 +1173,13 @@ def ____gemm_code(check_ab, a_init, b_init):
         return 0;  //success!
 
         _dot_execute_fallback:
-        PyErr_SetString(PyExc_NotImplementedError, 
+        PyErr_SetString(PyExc_NotImplementedError,
             "dot->execute() fallback");
         return -1;
 
         _dot_execute_fail:
         if (error_string == NULL)
-            PyErr_SetString(PyExc_ValueError, 
+            PyErr_SetString(PyExc_ValueError,
                 "dot->execute() cant run on these inputs");
         return -1;
 
